@@ -1,15 +1,13 @@
 from . import idnadata
+import bisect
 import unicodedata
 import re
 import sys
+from .intranges import intranges_contain
 
 _virama_combining_class = 9
 _alabel_prefix = b'xn--'
-_unicode_dots_re = re.compile(u'[\u002e\u3002\uff0e\uff61]')
-
-if sys.version_info[0] == 3:
-    unicode = str
-    unichr = chr
+_unicode_dots_re = re.compile('[\u002e\u3002\uff0e\uff61]')
 
 class IDNAError(UnicodeError):
     """ Base exception for all IDNA-encoding related problems """
@@ -32,16 +30,20 @@ class InvalidCodepointContext(IDNAError):
 
 
 def _combining_class(cp):
-    return unicodedata.combining(unichr(cp))
+    v = unicodedata.combining(chr(cp))
+    if v == 0:
+        if not unicodedata.name(chr(cp)):
+            raise ValueError('Unknown character in unicodedata')
+    return v
 
 def _is_script(cp, script):
-    return ord(cp) in idnadata.scripts[script]
+    return intranges_contain(ord(cp), idnadata.scripts[script])
 
 def _punycode(s):
     return s.encode('punycode')
 
 def _unot(s):
-    return 'U+{0:04X}'.format(s)
+    return 'U+{:04X}'.format(s)
 
 
 def valid_label_length(label):
@@ -51,9 +53,9 @@ def valid_label_length(label):
     return True
 
 
-def valid_string_length(label):
+def valid_string_length(label, trailing_dot):
 
-    if len(label) > 254:
+    if len(label) > (254 if trailing_dot else 253):
         return False
     return True
 
@@ -69,7 +71,6 @@ def check_bidi(label, check_ltr=False):
             raise IDNABidiError('Unknown directionality in label {} at position {}'.format(repr(label), idx))
         if direction in ['R', 'AL', 'AN']:
             bidi_label = True
-            break
     if not bidi_label and not check_ltr:
         return True
 
@@ -154,9 +155,9 @@ def valid_contextj(label, pos):
         ok = False
         for i in range(pos-1, -1, -1):
             joining_type = idnadata.joining_types.get(ord(label[i]))
-            if joining_type == 'T':
+            if joining_type == ord('T'):
                 continue
-            if joining_type in ['L', 'D']:
+            if joining_type in [ord('L'), ord('D')]:
                 ok = True
                 break
 
@@ -166,9 +167,9 @@ def valid_contextj(label, pos):
         ok = False
         for i in range(pos+1, len(label)):
             joining_type = idnadata.joining_types.get(ord(label[i]))
-            if joining_type == 'T':
+            if joining_type == ord('T'):
                 continue
-            if joining_type in ['R', 'D']:
+            if joining_type in [ord('R'), ord('D')]:
                 ok = True
                 break
         return ok
@@ -207,11 +208,11 @@ def valid_contexto(label, pos, exception=False):
 
     elif cp_value == 0x30fb:
         for cp in label:
-            if cp == u'\u30fb':
+            if cp == '\u30fb':
                 continue
-            if not _is_script(cp, 'Hiragana') and not _is_script(cp, 'Katakana') and not _is_script(cp, 'Han'):
-                return False
-        return True
+            if _is_script(cp, 'Hiragana') or _is_script(cp, 'Katakana') or _is_script(cp, 'Han'):
+                return True
+        return False
 
     elif 0x660 <= cp_value <= 0x669:
         for cp in label:
@@ -239,14 +240,19 @@ def check_label(label):
 
     for (pos, cp) in enumerate(label):
         cp_value = ord(cp)
-        if cp_value in idnadata.codepoint_classes['PVALID']:
+        if intranges_contain(cp_value, idnadata.codepoint_classes['PVALID']):
             continue
-        elif cp_value in idnadata.codepoint_classes['CONTEXTJ']:
-            if not valid_contextj(label, pos):
-                raise InvalidCodepointContext('Joiner {} not allowed at position {} in {}'.format(_unot(cp_value), pos+1, repr(label)))
-        elif cp_value in idnadata.codepoint_classes['CONTEXTO']:
+        elif intranges_contain(cp_value, idnadata.codepoint_classes['CONTEXTJ']):
+            try:
+                if not valid_contextj(label, pos):
+                    raise InvalidCodepointContext('Joiner {} not allowed at position {} in {}'.format(
+                        _unot(cp_value), pos+1, repr(label)))
+            except ValueError:
+                raise IDNAError('Unknown codepoint adjacent to joiner {} at position {} in {}'.format(
+                    _unot(cp_value), pos+1, repr(label)))
+        elif intranges_contain(cp_value, idnadata.codepoint_classes['CONTEXTO']):
             if not valid_contexto(label, pos):
-                raise InvalidCodepointContext('Codepoint {} not allowed at position {}'.format(_unot(cp_value), pos+1, repr(label)))
+                raise InvalidCodepointContext('Codepoint {} not allowed at position {} in {}'.format(_unot(cp_value), pos+1, repr(label)))
         else:
             raise InvalidCodepoint('Codepoint {} at position {} of {} not allowed'.format(_unot(cp_value), pos+1, repr(label)))
 
@@ -257,18 +263,17 @@ def alabel(label):
 
     try:
         label = label.encode('ascii')
-        try:
-            ulabel(label)
-        except:
-            raise IDNAError('The label {} is not a valid A-label'.format(label))
+        ulabel(label)
+        if not valid_label_length(label):
+            raise IDNAError('Label too long')
         return label
-    except UnicodeError:
+    except UnicodeEncodeError:
         pass
 
     if not label:
         raise IDNAError('No Input')
 
-    label = unicode(label)
+    label = str(label)
     check_label(label)
     label = _punycode(label)
     label = _alabel_prefix + label
@@ -284,13 +289,17 @@ def ulabel(label):
     if not isinstance(label, (bytes, bytearray)):
         try:
             label = label.encode('ascii')
-        except UnicodeError:
+        except UnicodeEncodeError:
             check_label(label)
             return label
 
     label = label.lower()
     if label.startswith(_alabel_prefix):
         label = label[len(_alabel_prefix):]
+        if not label:
+            raise IDNAError('Malformed A-label, no Punycode eligible content found')
+        if label.decode('ascii')[-1] == '-':
+            raise IDNAError('A-label must not end with a hyphen')
     else:
         check_label(label)
         return label.decode('ascii')
@@ -300,44 +309,88 @@ def ulabel(label):
     return label
 
 
-def encode(s, strict=False):
+def uts46_remap(domain, std3_rules=True, transitional=False):
+    """Re-map the characters in the string according to UTS46 processing."""
+    from .uts46data import uts46data
+    output = ''
+    try:
+        for pos, char in enumerate(domain):
+            code_point = ord(char)
+            uts46row = uts46data[code_point if code_point < 256 else
+                bisect.bisect_left(uts46data, (code_point, 'Z')) - 1]
+            status = uts46row[1]
+            replacement = uts46row[2] if len(uts46row) == 3 else None
+            if (status == 'V' or
+                    (status == 'D' and not transitional) or
+                    (status == '3' and not std3_rules and replacement is None)):
+                output += char
+            elif replacement is not None and (status == 'M' or
+                    (status == '3' and not std3_rules) or
+                    (status == 'D' and transitional)):
+                output += replacement
+            elif status != 'I':
+                raise IndexError()
+        return unicodedata.normalize('NFC', output)
+    except IndexError:
+        raise InvalidCodepoint(
+            'Codepoint {} not allowed at position {} in {}'.format(
+            _unot(code_point), pos + 1, repr(domain)))
 
+
+def encode(s, strict=False, uts46=False, std3_rules=False, transitional=False):
+
+    if isinstance(s, (bytes, bytearray)):
+        s = s.decode('ascii')
+    if uts46:
+        s = uts46_remap(s, std3_rules, transitional)
     trailing_dot = False
     result = []
     if strict:
         labels = s.split('.')
     else:
         labels = _unicode_dots_re.split(s)
-    if labels[-1] == '':
-        labels = labels[0:-1]
-        trailing_dot = True
-    if not labels:
+    if not labels or labels == ['']:
         raise IDNAError('Empty domain')
+    if labels[-1] == '':
+        del labels[-1]
+        trailing_dot = True
     for label in labels:
-        result.append(alabel(label))
+        s = alabel(label)
+        if s:
+            result.append(s)
+        else:
+            raise IDNAError('Empty label')
     if trailing_dot:
         result.append(b'')
-    return b'.'.join(result)
+    s = b'.'.join(result)
+    if not valid_string_length(s, trailing_dot):
+        raise IDNAError('Domain too long')
+    return s
 
 
-def decode(s, strict=False):
+def decode(s, strict=False, uts46=False, std3_rules=False):
 
+    if isinstance(s, (bytes, bytearray)):
+        s = s.decode('ascii')
+    if uts46:
+        s = uts46_remap(s, std3_rules, False)
     trailing_dot = False
     result = []
-    if isinstance(s, (bytes, bytearray)):
-        labels = s.split(b'.')
+    if not strict:
+        labels = _unicode_dots_re.split(s)
     else:
-        if not strict:
-            labels = _unicode_dots_re.split(s)
-        else:
-            labels = s.split(u'.')
-    if not labels[-1]:
-        labels = labels[0:-1]
-        trailing_dot = True
-    if not labels:
+        labels = s.split('.')
+    if not labels or labels == ['']:
         raise IDNAError('Empty domain')
+    if not labels[-1]:
+        del labels[-1]
+        trailing_dot = True
     for label in labels:
-        result.append(ulabel(label))
+        s = ulabel(label)
+        if s:
+            result.append(s)
+        else:
+            raise IDNAError('Empty label')
     if trailing_dot:
-        result.append(u'')
-    return u'.'.join(result)
+        result.append('')
+    return '.'.join(result)
